@@ -10,109 +10,215 @@ use 5.006;
 use strict;
 use version;
 
-use Cwd                     qw( cwd             );
-use File::Basename          qw( basename        );
-use File::Copy::Recursive   qw( dircopy         );
-use File::Find              qw( find            );
-use FindBin                 qw( $Bin            );
-use List::Util              qw( first           );
-use Symbol                  qw( qualify_to_ref  );
+use NEXT;
+
+use Cwd                     qw( cwd                     );
+use File::Basename          qw( basename                );
+use File::Copy::Recursive   qw( dircopy                 );
+use File::Find              qw( find                    );
+use FindBin                 qw( $Bin                    );
+use List::Util              qw( first                   );
+use Scalar::Util            qw( blessed                 );
+use Symbol                  qw( qualify qualify_to_ref  );
+
+use Module::FromPerlVer::Dir;
 
 ########################################################################
 # package variables & sanity checks
 ########################################################################
 
-our $VERSION    = '0.0.1';
+our $VERSION    = version->parse( 'v0.0.2' )->numify;
 
-my $default_d   = 'version';
+my %defaultz = 
+(
+    # pick your poison
+
+    default_use     => 'dir',
+    use_dir         => '',
+    use_git         => '',
+
+    # global settings
+
+    perl_version    => '',
+    version_from    => '',
+    verbose         => '',
+
+    # directory
+
+    version_dir     => 'version',
+
+    # git
+
+    git_prefix      => 'perl/',
+);
 
 ########################################################################
 # utility subs
 ########################################################################
 
+my $extract_args
+= sub
+{
+    my ( undef, %argz ) = @_;
+
+    while( my ($k, $def_v) = each %defaultz )
+    {
+        my $env_k   = uc $k;
+        my $env_v   = $ENV{ $env_k };
+
+        for( $argz{ $k } )
+        {
+            if( defined $env_v )
+            {
+                print "# Env: $k = '$env_v' ($env_k)";
+
+                $argz{ $k } = $env_v;
+            }
+            elsif( defined $argz{ $k } )
+            {
+                print "# Arg: $k = '$_'";
+            }
+            elsif( $def_v )
+            {
+                print "# Def: $k = '$def_v'";
+
+                $argz{ $k } = $def_v;
+            }
+            else
+            {
+                # nothing more to do: the default
+                # isn't overridden and it false,
+                # no argument supplied: skip it.
+            }
+        }
+    }
+
+    \%argz
+};
+
+my $make_extract
+= sub
+{
+    my $argz    = shift;
+
+    my @found 
+    = map
+    {
+        m{ (use_ ([a-z]+) ) }x && delete $argz->{ $1 }
+        ? $2
+        : ()
+    }
+    keys %$argz;
+
+    local $, = ' ';
+
+    if( @found > 1 )
+    {
+        die "Invalid: Multiple sources, @found\n";
+    }
+    elsif( @found )
+    {
+        # nothing more do do
+    }
+    else
+    {
+        push @found, delete $argz->{ default_use };
+    }
+
+    my $type    = ucfirst $found[0];
+    my $pkg     = qualify $type;
+
+    $argz->{ type } = $type;
+
+    print "# Extract with: '$type' ($pkg)";
+
+    $pkg->new( $argz )
+};
+
+my $extract_perl_v
+= sub
+{
+    my $extract = shift;
+    my $perl_v  = '';
+
+    if( $perl_v = $extract->value( 'perl_version' ) )
+    {
+        eval
+        {
+            version->parse( $perl_v )->numify
+        }
+        or 
+        die "Unparsable perl version: '$perl_v'\n"
+    }
+    elsif( my $path = $extract->{ version_from } )
+    {
+        -e $path or die "Bogus version_from: non-existant '$path'\n";
+        -f _     or die "Bogus version_from: non-file     '$path'\n";
+        -s _     or die "Bogus version_from: zero-size    '$path'\n";
+        -r _     or die "Bogus version_from: non-readable '$path'\n";
+
+        # at this point the version file seems minimally usable.
+
+        open my $fh, '<', $path
+        or
+        die "Botched version_from: open '$path', $!\n";
+
+        $perl_v
+        = first
+        {
+            if
+            (
+                my ( $min_v )
+                = m{ \buse \s+ (v? 5[.] +?) \s* ; }
+            )
+            {
+                $perl_v 
+                = version->parse( $min_v )->numify
+                or 
+                die "Invalid version string: '$_' ($path)"
+            }
+            elsif
+            (
+                my ( $max_v ) 
+                = m{ \bno  \s+ (v? 5[.] +?) \s* ; }
+            )
+            {
+                $perl_v
+                = version->parse( $max_v )->numify - 0.000001
+                or 
+                die "Invalid version string: '$_' ($path)"
+            }
+            else
+            {
+                ''
+            }
+        }
+        readline $fh
+        or
+        die "Bogus version_from: '$path' lacks '(use|no) version'.\n";
+    }
+    else
+    {
+        # this *should* always be parseable.
+
+        $perl_v = version->parse( $^V )->numify
+        or
+        die "Severe weirdness: unparsable \$^V ($^V).\n";
+    }
+
+    $extract->value( perl_version => $perl_v )
+};
+
 my @handlerz = 
 (
     sub
     {
-        # set running perl version.
+        # acquire running perl version.
 
-        my $argz    = shift;
-        my $path    = $argz->{ version_from } || '';
-        my $perl_v  = '';
+        my $extract = shift;
+        my $perl_v  = $extract->$extract_perl_v;
 
-        if( my $value = $ENV{ COMPATIBLE_VERSION } )
-        {
-            warn "Overriding '$path' with '$value' from COMPATIBLE_VERSION.\n"
-            if $path;
-
-            $perl_v = version->parse( $value )->numify
-            or
-            die "Botched version_from: invalid '$value' from PERL_VERSION.\n";
-        }
-        elsif( $path )
-        {
-            -e $path or die "Bogus version_from: non-existant '$path'\n";
-            -f _     or die "Bogus version_from: non-file     '$path'\n";
-            -s _     or die "Bogus version_from: zero-size    '$path'\n";
-            -r _     or die "Bogus version_from: non-readable '$path'\n";
-
-            # at this point the version file seems minimally usable.
-
-            my $found
-            = do
-            {
-                open my $fh, '<', $path
-                or
-                die "Botched version_from: open '$path', $!\n";
-
-                first
-                {
-                    if
-                    (
-                        my ( $min_v )
-                        = m{ \buse \s+ (v? 5[.] +?) \s* ; }
-                    )
-                    {
-                        $perl_v 
-                        = version->parse( $min_v )->numify
-                        or 
-                        die "Invalid version string: '$_' ($path)"
-                    }
-                    elsif
-                    (
-                        my ( $max_v ) 
-                        = m{ \bno  \s+ (v? 5[.] +?) \s* ; }
-                    )
-                    {
-                        $perl_v
-                        = version->parse( $max_v )->numify - 0.000001
-                        or 
-                        die "Invalid version string: no '$_' ($path)"
-                    }
-                    else
-                    {
-                        ''
-                    }
-                }
-                readline $fh
-                or
-                die "Bogus version_from: '$path' lacks 'use version'.\n";
-            };
-        }
-        else
-        {
-            # this *should* to be parseable.
-
-            $perl_v = version->parse( $^V )->numify
-            or
-            die "Severe weirdness: unparsable \$^V ($^V).\n";
-
-        }
-
-        # belt & suspenders time
-
-        $perl_v
-        or die "Broken perl_version: no version extracted.";
+        # lacking an exception, we have a perl version.
 
         *{ qualify_to_ref 'perl_version' }
         = sub 
@@ -120,74 +226,45 @@ my @handlerz =
             $perl_v
         };
 
-        print "# Perl version: '$perl_v'";
+        print '# Perl version: ' . perl_version();
 
         return
     },
 
     sub
     {
-        # exract version parent directory
+        # extract source of the inputs. this will be
+        # either the ./version directory or a git tag
+        # prefix.
 
-        my $argz    = shift;
-        my $dir     = $argz->{ version_dir } || $default_d;
+        my $extract = shift;
+        my $prefix  = $extract->source_prefix;
 
-        my $path
-        = first
+        *{ qualify_to_ref 'source_prefix' }
+        = sub
         {
-            -e 
-        }
-        (
-            $dir        ,
-            "$Bin/$dir" ,
-            "./$dir"    ,
-        )
-        or die "Bogus version_dir: Non-existant: '$dir' ($Bin)";
+            $prefix
+        };
 
-        for my $cwd ( cwd )
-        {
-            # convert $path to relative.
-
-            my $i   = length $cwd;
-
-            index $path, "$cwd/"
-            or
-            substr $path, 0, $i, '.'
-        }
-
-        -e $path    or die "Bogus version_dir: non-existant '$path'";
-        -d _        or die "Bogus version_dir: non-dir      '$path'";
-        -r _        or die "Bogus version_dir: non-readable '$path'";
-        -x _        or die "Bogus version_dir: non-execable '$path'";
-
-        my @found   = glob "$path/*"
-        or die "Botchd version_dir: '$path' is empty directory.\n";
-
-        *{ qualify_to_ref 'version_dir' }
-        = sub { $path };
-
-        print "# Version directory: '$path'";
+        print '# Source prefix: ' . source_prefix();
 
         return
     },
 
     sub
     {
-        # locate source dir supporting perl version.
+        # locate source appropriate for perl version.
 
-        my $perl_v      = perl_version();
-        my $version_d   = version_dir();
+        my $extract = shift;
+        my $perl_v  = perl_version();
 
-        my $source_d
-        = first
+        my @sourcz  = $extract->module_sources
+        or die "Botched module_sources: empty source list.\n";
+
+        my $found
+        =first
         {
-            $_
-        }
-        map
-        {
-            $perl_v >= $_->[0] 
-            ? $_->[1]
-            : ()
+            $perl_v >= $_->[0]
         }
         sort
         {
@@ -195,21 +272,33 @@ my @handlerz =
         }
         map
         {
-            my $dir_v   
-            = version->parse( basename $_ )->numify;
-
-            [ $dir_v, $_ ]
+            [
+                version->parse( basename $_ )->numify,
+                $_
+            ]
         }
-        glob "$version_d/*"
-        ;
-
-        *{ qualify_to_ref 'source_dir' }
-        = sub
+        @sourcz
+        or do
         {
-            $source_d
+            local $,    = "\n#\t";
+
+            die 'Botched module_source: no appropriate version',
+            @sourcz, "\n";
         };
 
-        print "# Source directory:  '$source_d'";
+        my ( $mod_v, $path ) = @$found;
+
+        print "# Module version: $mod_v <= $perl_v";
+
+        $extract->value( module_source => $path );
+
+        *{ qualify_to_ref 'module_source' }
+        = sub
+        {
+            $path
+        };
+
+        print '# Module source: ' . module_source();
 
         return
     },
@@ -221,98 +310,89 @@ my @handlerz =
         # pre-assigning the hashref's simplifies dealing
         # with a collection of empty dir's.
 
-        my $source_d    = source_dir();
-        my @pathz       = ( [], [] );
+        my $extract = shift;
+        my $ref     = qualify_to_ref 'source_files';
 
-        my $n           = length $source_d;
-
-        find
-        sub
+        if( my @pathz   = $extract->source_files )
         {
-            my $path    = $File::Find::name;
-
-            $path ne $source_d
-            or return;
-
-            my $rel     = '.' . substr $path, $n;
-
-            my $i
-            = -d $_
-            ? 1
-            : 0
-            ;
-
-            push @{ $pathz[ $i ] }, $rel;
-        },
-        $source_d;
-
-        # deal with a set of empty dirs.
-
-        @{ $pathz[0] }
-        or warn "No input files found: '$source_d'";
-
-        *{ qualify_to_ref 'source_files' }
-        = sub
-        {
-            wantarray
-            ? @pathz
-            : $pathz[0]
-        };
-
-        local $,    = "\n#\t";
-        print '# Source files:', @{ $pathz[0] };
-
-        return
-    },
-
-    sub
-    {
-        *{ qualify_to_ref 'cleanup' }
-        = sub
-        {
-            my ( $filz, $dirz ) = source_files();
-
-            unlink @$filz;
-
-            -e $_ && warn "Failed removal: '$_'"
-            for @$filz;
-
-            # i.e., don't clobber dir's which have any
-            # files we didn't copy in them. this is most
-            # likely for ./t which can have common tests.
-
-            for my $dir ( @$dirz )
+            *$ref
+            = sub
             {
-                ( my @a = glob "$dir/*" )
-                or 
-                rmdir $dir;
-            }
-        };
+                wantarray
+                ? @pathz
+                : $pathz[0]
+            };
+
+            local $,    = "\n#\t";
+            print '# Source files:', @{ scalar source_files() };
+        }
+        else
+        {
+            *$ref
+            = sub
+            {
+                # avoid returning true in scalar context
+
+                return
+            };
+
+            print '# No source files.';
+        }
 
         return
     },
 
     sub
     {
-        my ( $filz, $dirz ) = source_files();
-        my $expect  = 1 + @$filz + @$dirz;
+        my $ref = qualify_to_ref 'cleanup';
 
-        *{ qualify_to_ref 'copy_source_dir' }
+        if( my ( $filz, $dirz ) = source_files() )
+        {
+            *$ref
+            = sub
+            {
+                unlink @$filz;
+
+                -e $_ && warn "Failed removal: '$_'"
+                for @$filz;
+
+                # i.e., don't clobber dir's which have any
+                # files we didn't copy in them. this is most
+                # likely for ./t which can have common tests.
+
+                for my $dir ( @$dirz )
+                {
+                    ( my @a = glob "$dir/*" )
+                    or 
+                    rmdir $dir;
+                }
+
+                return
+            };
+
+            print '# Install cleanup.';
+        }
+        else
+        {
+            *$ref   = sub{};
+
+            print '# Nothing to clean up.';
+        }
+
+        return
+    },
+
+    sub
+    {
+        my $extract = shift;
+
+        *{ qualify_to_ref 'get_files' }
         = sub
         {
-            my $dir     = source_dir();
-            my $found   = dircopy $dir, '.';
-            
-            print "# Copied: $found files from '$dir'";
-
-            $found != $expect
-            and
-            print "# Oddity: mismatched count $found != $expect.";
-
-            $found
+            $extract->get_files
         };
 
-        print "# Expect: $expect files to be copied.";
+        print "# Get files with: '$extract->{ type }'";
 
         return
     }
@@ -326,23 +406,18 @@ sub import
 {
     local $\    = "\n";
 
-    my ( undef, %argz ) = @_;
+    my $argz    = &$extract_args;
+    my $extract = $make_extract->( $argz );
 
-    for my $sub ( @handlerz )
-    {
-        $sub->( \%argz );
-    }
+    $extract->$_ for @handlerz;
 
     # at this point the args are consumed and if we are 
-    # still alive the copy function has been installed.
+    # still alive the utility subs are installed.
     #
-    # caller can use cleanup() to remove files copied or 
-    # source_files() to get a list of them.
-    #
-    # this returns the count from File::Copy::Recursive,
-    # which should be true if anything was found to copy.
+    # Note: this returns true if the handler cycle completed
+    # (either no_copy value or result of making the copy).
 
-    $argz{ no_copy }
+    $extract->{ no_copy }
     or
     copy_source_dir()
 }
@@ -399,9 +474,15 @@ Module::FromPerlVer - install modules compatible with the running perl.
 
     # this can be useful for regression testing earlier 
     # versions of modules against the running version 
-    # (e.g., validating experimental features).
+    # (e.g., validating experimental features, benchmarking
+    # older code against newer perl executables).
+    #
+    # arguments can be passed to use or set in the environment
+    # with a variable of the option uppercased. For example:
 
-    $ COMPATIBLE_VERSION='5.024002' perl Makefile.PL;
+    $ PERL_VERSION='5.024002' perl Makefile.PL;
+
+    use Module::FromPerlVer qw( perl_version v5.24.1 );
 
     # use a file to determine the perl version.
     # this takes the use-ed version from the file
@@ -410,6 +491,8 @@ Module::FromPerlVer - install modules compatible with the running perl.
 
     use Module::FromPerlVer   qw( version_from ./lib/foobar.pm );
 
+    VERSION_FROM='./lib/foobar.pm' use Module::FromPerlVer;
+
     # or a one-line datafile.
     # "use vX.Y.Z" will find up to that version, "no vX.Y.Z" will 
     # use up to that version minus '0.000001'.
@@ -421,14 +504,22 @@ Module::FromPerlVer - install modules compatible with the running perl.
 
     echo 'no v5.24.2' > ./perl-version;
 
-    use Module::FromPerlVer   qw( version_from perl-version );
+    use Module::FromPerlVer   qw( version_from file_path );
 
     # override the source directory:
 
     # maybe you don't like the name 'version', you
     # prefer "history" instead.
 
-    use Module::FromPerlVer   qw( version_dir history );
+    use Module::FromPerlVer   qw( source_prefix history );
+
+    # maybe you don't like storing your older versions in 
+    # a directory, you prefer to use git with tags to snag
+    # the working directory with a default tag prefix of 
+    # "perl/" (or pick your own):
+
+    use Module::FromPerlVer qw( use_git 1 );
+    use Module::FromPerlVer qw( source_prefix ThingyDotNet/ );
 
     # set up everything but skip making the copy.
     #
@@ -440,14 +531,14 @@ Module::FromPerlVer - install modules compatible with the running perl.
 
     use Module::FromPerlVer   qw( no_copy 1 );
 
-
     my ( $filz, $dirz ) = Module::FromPerlVer::source_files;
 
     Module::FromPerlVer->copy_source_dir;
 
     my $madness = 'Module::FromPerlVer';
+    my $method  = 'cleanup';
 
-    $madness->cleanup;
+    $madness->$method;
 
 
 
@@ -463,8 +554,11 @@ for $^V are necessary.
 
 =head2 Using the module
 
-Using this module requires two things: The module and a local
-filesysem labeled by compatible Perl version.
+Using this module requires two things: The module and a source
+for perl-version-speific files. With "use_dir" the files are
+found in a specific local directory; with "use_git" the are
+found via (guess... if you said "cvs checkout from ./attic"
+you'd be wrong).
 
 =over 4
 
@@ -486,11 +580,24 @@ files are suitable for that version of perl. Common
 examples are README, MANIFEST, Changes, ./lib, ./t,
 ./bin.
 
+=item Version Tag
+
+With git the current repository is scanned via "git tags"
+with the source prefix. The versions need to look path-ish:
+
+    perl/v5.6.1
+    Perly/5.005003
+    MyPerlStuff/v5.24
+
+will all do nicely: they have a '/' with a "basename" of
+a parsable perly version.
+
 =item Basic Use
 
 Using the module in setup code (e.g., Makefile.PL):
 
-    use Devel::PerlCompatiable;
+    use Module::FromPerlVer;
+    use Module::FromPerlVer qw( use_dir 1 );
 
 will look in the ./version directory, sort the subdir's 
 in numerically decreasing order by version (see version)
@@ -519,7 +626,7 @@ module's compatible directory are:
 
 =over 4
 
-=item $ENV{ COMPATIBLE_VERSION }
+=item $ENV{ PERL_VERSION }
 
 Any true value of  will be used. If this is supplied with
 the "version_from" argument a warning will be issued and
@@ -570,7 +677,7 @@ with a specific release of the module.
 
 =item Overriding "version".
 
-Say you hate the name version:
+If you hate the name "version", use "version_dir" to set it:
 
     use Module::FromPerlVer qw( version_dir history );
 
@@ -589,6 +696,58 @@ below) but will not execute copy_source_dir().
 
 =back
 
+=head2 Use Messages
+
+The module supplies most of the information necessary to 
+debug environment and configuration issues when import()
+is called:
+
+    # Def: default_use = 'dir'
+    # Def: version_dir = 'version'
+    # Def: git_prefix = 'perl/'
+    # Extract with: 'Dir' (Module::FromPerlVer::Dir)
+    # Perl version: 5.026001
+    # Source prefix: ./t/version
+    # Module version: 5.005003 <= 5.026001
+    # Module source: ./t/version/5.005003
+    # Source files:
+    #   ./lib/Foo.pm
+    #   ./pod/Foo.pod
+    #   ./etc/version.dat
+    # Install cleanup.
+    # Get files with: 'Dir'
+
+
+=over 4
+
+=item Configuration
+
+    # Def: default_use = 'dir'
+    # Def: version_dir = 'version'
+    # Def: git_prefix = 'perl/'
+
+There are three variants: 'Def', 'Env', 'Arg' which 
+show where the values came from. 
+
+=item Handler
+
+This shows which of use_* or the default_use was
+chosen for the copy handler. In this case the 
+default_use value of 'dir' provided the source:
+
+    # Extract with: 'Dir' (Module::FromPerlVer::Dir)
+
+=item Utility Installation
+
+When the utility subs are installed they provide 
+basic information about the values selected from
+the arguments, perl version, and workspace.
+
+See "Utility Subs" (below) for more information on
+what these mean.
+
+=back
+
 =head2 Utility Subs
 
 A few utility sub's are installed into Module::FromPerlVer 
@@ -631,22 +790,28 @@ paths with the identical effect:
 Returns the parsed, numified Perl version value (see 
 version module).
 
-=item version_dir()
+=item version_source()
 
-Returns the basename of the version directory
-(i.e., default 'version').
+Returns the source of the version directory
+(e.g., dir "version" or tag "perl/").
 
-=item source_dir()
+=item version_source()
 
-The subdir from which the files are sourced. This will be a
-relative path from $FindBin::Bin, including the version 
-directory:
+The subdir or tag from which the files are sourced. This will 
+be a relative path from $FindBin::Bin, including the 
+version directory:
 
     ./version/5.006001
     ./version/v5.24.2
 
+    perl/v5.24
+    perl5.005_003
+
 =item source_files()
 
+For directory source only (i.e., where the files can be 
+looked up easily beforehand):
+
 Used in a scalar context this returns an arrayref of relative
 paths to files copied from the source_dir into the working 
 directory by copy_files().
@@ -656,9 +821,9 @@ files one of the directories. The former is used with unlink
 in cleanup to remove only the files that were copied; the 
 latter is used to rmdir empty directories.
 
-=item copy_source_dir()
+=item get_files()
 
-This executes the copy of all source files into '.'.
+This executes the copy/checkout of source files.
 
 Note: If import is called with "no_copy" and a true
 value then this will have to be called, see examples
@@ -666,16 +831,35 @@ for calling convention, above.
 
 =item cleanup()
 
-This first executes an unlink the files, then walks the 
-dir's executing rmdir on the empty ones.
+For Dir this removes the files copied, empty dir's left behind. This 
+approach allows for some files to be re-used across multiple releases 
+without getting clobbered by cleanup.  Likely examples are common 
+tests (i.e., t/*.t) or a MANIFEST for modules which have all of the 
+same files in each version.
 
-This approach allows for some files to be re-used across 
-multiple releases without getting clobbered by cleanup.
-Likely examples are common tests (i.e., t/*.t) or a MANIFEST
-for modules which have all of the same files in each version.
+For Git this checks out whatever branch was in use when copy_files() 
+was called. That's the best guess I have for what was in use at the
+time a verion-specific tag was checked out.
 
 =back
 
+=head1 NOTES
+
+=over 4
+
+=item Q: Why tags, not branches?
+
+A given brach may have un-tested updates on it. At that point it
+is not safe to assume HEAD will always work; at *that* point there
+isn't any good way to automatically checkout a given branch.
+
+Tags are static, can be assigned by the developer for specific
+commits, and can be re-assigned as necessary if code is updated.
+
+This also saves a whole lot of separate branches for version-
+specific generations of code. Just leave them in master -- or
+wherever you work -- with the tags behind.
+
 =back
 
 =head1 SEE ALSO
